from django.apps import apps
from django.core.exceptions import ImproperlyConfigured


def register(model):
    """
    Register the given model(s) classes and wrapped ModelSite class:
    @register(Author)
    class AuthorSite(superadmin.ModelSite):
        pass
    """
    from superadmin import ModelSite, site

    def _model_site_wrapper(site_class):
        if not model:
            raise ValueError("One model must be passed to register.")

        model_class = None
        if isinstance(model, str):
            try:
                app_name, model_name = model.split(".")
                model_class = apps.get_model(app_name, model_name)
            except ValueError:
                raise ImproperlyConfigured("Does not exist '%s' model" % model)

        if not issubclass(site_class, ModelSite):
            raise ValueError("Wrapped class must subclass ModelSite.")

        site.register(model_class if model_class else model, site_class=site_class)

        return site_class

    return _model_site_wrapper



# Generated by Django 3.1.1 on 2020-11-28 15:07

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("superadmin", "unaccent_extension"),
    ]

    operations = [
        migrations.CreateModel(
            name="Action",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "to",
                    models.PositiveSmallIntegerField(
                        choices=[(1, "Modelo"), (2, "Vista")],
                        verbose_name="acción hacia un",
                    ),
                ),
                (
                    "app_label",
                    models.CharField(max_length=128, verbose_name="aplicación"),
                ),
                (
                    "name",
                    models.CharField(
                        max_length=128, verbose_name="nombre de la acción"
                    ),
                ),
                (
                    "element",
                    models.CharField(max_length=128, verbose_name="elemento accionado"),
                ),
                (
                    "permissions",
                    models.ManyToManyField(
                        blank=True,
                        to="auth.Permission",
                        verbose_name="permisos específicos",
                    ),
                ),
            ],
            options={
                "verbose_name": "acción",
                "verbose_name_plural": "acciones",
                "ordering": ("name",),
                "unique_together": {("app_label", "element")},
            },
        ),
        migrations.CreateModel(
            name="Menu",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128, verbose_name="nombre")),
                (
                    "route",
                    models.CharField(
                        editable=False,
                        max_length=512,
                        unique=True,
                        verbose_name="ruta de acceso",
                    ),
                ),
                (
                    "icon_class",
                    models.CharField(
                        blank=True,
                        max_length=128,
                        null=True,
                        verbose_name="clase css del ícono",
                    ),
                ),
                (
                    "is_group",
                    models.BooleanField(
                        default=False, editable=False, verbose_name="agrupa"
                    ),
                ),
                (
                    "sequence",
                    models.PositiveSmallIntegerField(verbose_name="secuencia"),
                ),
                (
                    "is_active",
                    models.BooleanField(default=True, verbose_name="activo?"),
                ),
                (
                    "action",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="superadmin.action",
                        verbose_name="acción",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="submenus",
                        to="superadmin.menu",
                        verbose_name="menú padre",
                    ),
                ),
            ],
            options={
                "ordering": ("sequence"),
            },
        ),
    ]



""" Hydra init config """
from .sites import site
from .options import ModelSite
from django.utils.module_loading import autodiscover_modules

__all__ = ["site", "ModelSite"]

default_app_config = "superadmin.apps.SuperAdminConfig"


def autodiscover():
    autodiscover_modules("sites")



# Generated by Django 2.2.8 on 2019-12-10 17:20
from django.contrib.postgres.operations import UnaccentExtension
from django.db import migrations


class Migration(migrations.Migration):

    operations = [UnaccentExtension()]



# Python
import operator
from functools import reduce

# Django
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.forms.utils import pretty_name
from django.utils.html import format_html
from django.db.models import Q

from . import settings


class FieldService:
    FIELD_SEPARATOR = "__"
    LABEL_SEPARATOR = ":"

    @classmethod
    def get_field(cls, model, field):
        field = field.split(cls.LABEL_SEPARATOR)[0]
        names = field.split(cls.FIELD_SEPARATOR)
        name = names.pop(0)
        try:
            if len(names):
                related_model = model._meta.get_field(name).related_model
                if related_model:
                    return cls.get_field(related_model, cls.FIELD_SEPARATOR.join(names))
                else:
                    raise ImproperlyConfigured(f"The field <{name}> not is an object")
            field = model._meta.get_field(name)
            return field
        except FieldDoesNotExist:
            str_model = (
                model._meta.model_name if hasattr(model, "_meta") else str(model)
            )
            raise AttributeError(f"Does not exist attribute <{name}> for {str_model}")

    @classmethod
    def get_field_label(cls, model, field):
        try:
            name, verbose_name = field.split(cls.LABEL_SEPARATOR)
            return pretty_name(verbose_name)
        except ValueError:
            pass
        if "__str__" in field:
            label = str(model._meta.verbose_name)
            return pretty_name(label)
        names = field.split(cls.FIELD_SEPARATOR)
        name = names.pop(0)
        if not hasattr(model, name):
            str_model = (
                model._meta.model_name if hasattr(model, "_meta") else str(model)
            )
            raise AttributeError(f"Does not exist attribute <{name}> for {str_model}.")
        try:
            field = model._meta.get_field(name)
            if len(names):
                related_model = field.related_model
                return cls.get_field_label(
                    related_model, cls.FIELD_SEPARATOR.join(names)
                )
            label = field.verbose_name
        except FieldDoesNotExist:
            attr = getattr(object, name)
            if len(names):
                return cls.get_field_label(
                    attr(model) if callable(attr) else attr,
                    cls.FIELD_SEPARATOR.join(names),
                )
            label = name
        return pretty_name(label)

    @classmethod
    def get_field_value(cls, object, field):
        if "__str__" in field:
            return object
        if object is None:
            return object
        field = field.split(cls.LABEL_SEPARATOR)[0]
        names = field.split(cls.FIELD_SEPARATOR)
        name = names.pop(0)
        if not hasattr(object, name):
            raise AttributeError(
                f"Does not exist attribute <{name}> for {str(object)}."
            )
        if len(names):
            attr = getattr(object, name)
            return cls.get_field_value(
                attr() if callable(attr) else attr, cls.FIELD_SEPARATOR.join(names)
            )
        try:
            field = object._meta.get_field(name)
            if hasattr(field, "choices") and field.choices:
                return dict(field.choices).get(field.value_from_object(object))
            elif field.related_model:
                if field.one_to_many or field.many_to_many:
                    raise ImproperlyConfigured(
                        "OneToMany or ManyToMany is not supported: '%s' " % field.name
                    )
                try:
                    return field.related_model.objects.get(
                        pk=field.value_from_object(object)
                    )
                except field.related_model.DoesNotExist:
                    return None
            else:
                return field.value_from_object(object)
        except FieldDoesNotExist:
            attr = getattr(object, name)
            attr = attr() if callable(attr) else attr
            if isinstance(attr, bool):
                attr = settings.BOOLEAN_YES if attr else settings.BOOLEAN_NO
            return format_html(str(attr))

    @classmethod
    def get_field_type(cls, model, field):
        field = field.split(cls.LABEL_SEPARATOR)[0]
        names = field.split(cls.FIELD_SEPARATOR)
        name = names.pop(0)
        if not hasattr(model, name):
            str_model = (
                model._meta.model_name if hasattr(model, "_meta") else str(model)
            )
            raise AttributeError(f"Does not exist attribute <{name}> for {str_model}.")
        if len(names):
            if hasattr(model, "_meta"):
                return cls.get_field_type(
                    model._meta.get_field(name).related_model,
                    cls.FIELD_SEPARATOR.join(names),
                )
            else:
                attr = getattr(model, name)
                return cls.get_field_type(
                    attr() if callable(attr) else attr, cls.FIELD_SEPARATOR.join(names)
                )
        try:
            field = model._meta.get_field(name)
            type = model._meta.get_field(name).get_internal_type()
        except FieldDoesNotExist:
            type = "Function"
        return type


class FilterService:
    LOOKUPS = {
        "iexact": "Es igual a",
        "exact": "Es igual a",
        "iexact__exclude": "No es igual a",
        "exact__exclude": "No es igual a",
        "icontains": "Contiene",
        "contains": "Contiene",
        "icontains__exclude": "No contiene",
        "contains__exclude": "No contiene",
        "gte": "Mayor o igual que",
        "lte": "Menor o igual que",
    }

    EXCLUDE = "__exclude"

    @classmethod
    def get_lookup_label(cls, lookup):
        return cls.LOOKUPS.get(lookup)

    @classmethod
    def get_flatten_lookups(cls):
        return cls.LOOKUPS.keys()

    @classmethod
    def get_field_lookups(cls, model, field):
        field_lookups = []
        field = FieldService.get_field(model, field)
        if field.get_internal_type() == "BooleanField":
            field_lookups = [
                ("exact", cls.get_lookup_label("exact")),
                ("exact__exclude", cls.get_lookup_label("exact__exclude")),
            ]
            return field_lookups
        lookups = [
            ("iexact", "exact"),
            ("icontains", "contains"),
            ("gte",),
            ("lte",),
        ]
        allow_lookups = list(field.get_lookups().keys())
        for pair in lookups:
            for lookup in pair:
                if lookup in allow_lookups:
                    field_lookups.append((lookup, cls.get_lookup_label(lookup)))
                    if "exact" in lookup or "contains" in lookup:
                        field_lookups.append(
                            (
                                f"{lookup}__exclude",
                                cls.get_lookup_label(f"{lookup}__exclude"),
                            )
                        )
                    break
        return field_lookups

    @classmethod
    def get_choices(self, model, field):
        type = FieldService.get_field_type(model, field)
        field = FieldService.get_field(model, field)
        if type in ("ForeignKey", "OneToOneField", "ManyToManyField"):
            return field.related_model.objects.all()
        elif type == "BooleanField":
            choices = [(1, "Verdadero"), (0, "False")]
        elif field.choices:
            choices = list(field.choices)
        else:
            choices = []
        return choices

    @classmethod
    def get_previous_and_next(cls, queryset, instance):
        values = queryset.values_list("id", flat=True).all()
        for index, value in enumerate(values):
            if value == instance.pk:
                previous_id = None if index == 0 else values[index - 1]
                next_id = None if index == len(values) - 1 else values[index + 1]
                objects = queryset.filter(id__in=[previous_id, next_id])
                return {
                    "previous": previous_id if not previous_id else objects.first(),
                    "next": next_id if not next_id else objects.last(),
                    "current_index": index,
                    "total_entries": len(values),
                }

    @classmethod
    def get_params(self, model, session):
        app = model._meta.app_label
        model = model._meta.model_name
        filters = session.get("filters", [])
        params = {}
        for elem in filters:
            if elem["app"] == app and elem["model"] == model:
                params = elem["params"]
                break
        return params

    @classmethod
    def filter(cls, queryset, params):
        filter_query, exclude_query = map(
            lambda params: cls.get_query(params),
            cls.split_params(queryset.model, params),
        )
        if filter_query:
            queryset = queryset.filter(filter_query)
        if exclude_query:
            queryset = queryset.exclude(exclude_query)
        return queryset

    @classmethod
    def split_params(cls, model, params):
        lookup_params = {
            key: value for key, value in params.items() if cls.has_lookup(key)
        }
        filter_params = {
            key: value for key, value in lookup_params.items() if cls.EXCLUDE not in key
        }
        exclude_params = {
            key.split(cls.EXCLUDE)[0]: value
            for key, value in lookup_params.items()
            if cls.EXCLUDE in key
        }
        for params in [filter_params, exclude_params]:
            for key, value in params.items():
                lookup = cls.has_lookup(key)
                type = FieldService.get_field_type(model, key.split(f"__{lookup}")[0])
                if type == "BooleanField":
                    try:
                        value = bool(int(value))
                    except ValueError:
                        value = False
                params[key] = value
        return filter_params, exclude_params

    @classmethod
    def get_query(cls, params):
        args = []
        for field, value in params.items():
            args.append(Q(**{field: value}))
        if args:
            return reduce(operator.__and__, args)
        return args

    @classmethod
    def has_lookup(cls, key):
        for lookup in cls.get_flatten_lookups():
            if key.endswith(lookup):
                return lookup



""" Models for buid menus """

# Django
from django.db import models
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from django.utils.functional import cached_property
from django.utils.text import slugify
from django.apps import apps
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.conf import settings

# Local
from . import site
from .utils import import_class


class Action(models.Model):
    class ToChoices(models.IntegerChoices):
        MODEL = 1, "Modelo"
        CLASSVIEW = 2, "Vista"

    to = models.PositiveSmallIntegerField(
        choices=ToChoices.choices, verbose_name="acción hacia un"
    )
    app_label = models.CharField(max_length=128, verbose_name="aplicación")
    name = models.CharField(max_length=128, verbose_name="nombre de la acción")
    element = models.CharField(max_length=128, verbose_name="elemento accionado")
    permissions = models.ManyToManyField(
        Permission, blank=True, verbose_name="permisos específicos"
    )

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = "acción"
        verbose_name_plural = "acciones"
        unique_together = ("app_label", "element")
        ordering = ("name",)

    @property
    def model(self):
        if self.to != self.ToChoices.MODEL:
            return None
        try:
            model_class = apps.get_model(self.app_label, self.element)
            return model_class
        except LookupError:
            return None

    @property
    def view(self):
        if self.to != self.ToChoices.CLASSVIEW:
            return None
        try:
            app_config = apps.get_app_config(self.app_label)
            view_class = import_class(f"{app_config.name}.views", self.element)
            return view_class
        except (LookupError):
            return None

    def get_permissions(self):
        return [
            f"{perm.content_type.app_label}.{perm.codename}"
            for perm in self.cached_permissions
        ]

    @cached_property
    def cached_permissions(self):
        return list(self.permissions.select_related("content_type").all())

    def has_permissions(self, user, user_perms=None):
        if not user.is_authenticated or not user.is_active:
            return False

        if user.is_superuser:
            return True

        if user_perms is None:
            user_perms = set(user.get_all_permissions())

        if self.to == self.ToChoices.MODEL:
            basic_perms = any(
                f"{self.app_label}.{perm}_{self.element}" in user_perms
                for perm in ("view", "add", "change", "delete")
            )
        else:
            basic_perms = True

        specific_perms = all(p in user_perms for p in self.get_permissions())
        return basic_perms and specific_perms


class Menu(models.Model):
    """Models for menu"""

    parent = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="submenus",
        on_delete=models.CASCADE,
        verbose_name="menú padre",
    )
    name = models.CharField(max_length=128, verbose_name="nombre")
    route = models.CharField(
        max_length=512, unique=True, editable=False, verbose_name="ruta de acceso"
    )
    action = models.ForeignKey(Action, on_delete=models.CASCADE, verbose_name="acción")
    icon_class = models.CharField(
        max_length=128, blank=True, null=True, verbose_name="clase css del ícono"
    )
    is_group = models.BooleanField(default=False, editable=False, verbose_name="agrupa")
    sequence = models.PositiveSmallIntegerField(verbose_name="secuencia")
    is_active = models.BooleanField(default=True, verbose_name="activo?")

    class Meta:
        ordering = ("sequence",)

    def __str__(self):
        return f"{self.name} | {self.get_route()}"

    def get_route(self):
        route = (
            f"{self.parent.get_route()}/{slugify(self.name)}"
            if self.parent
            else slugify(self.name)
        )
        return route

    def get_url(self):
        url_name = None
        reverse_url = True
        if self.action.to == Action.ToChoices.MODEL:
            model_class = self.action.model
            if model_class and site.is_registered(model_class):
                model_site = site.get_modelsite(model_class)
                url_name = model_site.get_url_name("list")
        else:
            # url_name = f"site:{slugify(self.name)}"
            url_name = f"/{self.route}"
            reverse_url = False
        try:
            if reverse_url:
                url_name = reverse(url_name)
            return url_name
        except NoReverseMatch:
            if settings.DEBUG and url_name:
                print("DEBUG: Not found url for %s" % url_name)
        return "/page/not-found/"



""" Hydra Apps config  """

# Django
from django.apps import AppConfig


class SuperAdminConfig(AppConfig):
    name = "superadmin"

    def ready(self):
        from . import signals

        self.module.autodiscover()



# Generated by Django 3.1.13 on 2021-11-06 12:26

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("superadmin", "0001_action_menu"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="menu",
            options={"ordering": ("sequence",)},
        ),
    ]



""" Update View engine """
# Django
from django.views.generic import View
from django.views.generic import UpdateView as BaseUpdateView
from django.http import HttpResponseForbidden, JsonResponse

# Local
from .base import SiteView, get_base_view
from ..services import FilterService
from ..shortcuts import get_urls_of_site
from ..utils import import_all_mixins, import_mixin


class UpdateMixin:
    """Update View del modelo"""

    action = "update"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        params = FilterService.get_params(self.site.model, self.request.session)
        queryset = FilterService.filter(self.site.queryset, params)
        nav = FilterService.get_previous_and_next(queryset, self.object)
        if nav:
            nav["previous_url"] = (
                get_urls_of_site(self.site, nav["previous"])[self.action]
                if nav.get("previous")
                else None
            )
            nav["next_url"] = (
                get_urls_of_site(self.site, nav["next"])[self.action]
                if nav.get("next")
                else None
            )
        opts = {"nav": nav}
        if "site" in context:
            context["site"].update(opts)
        else:
            context.update({"site": opts})
        context.update(self.site.form_extra_context)
        return context

    def get_slug_field(self):
        return self.site.slug_field or super().get_slug_field()

    def get_success_url(self):
        urls = get_urls_of_site(self.site, object=self.object)
        return urls.get(self.site.update_success_url)


class UpdateView(SiteView):
    def view(self, request, *args, **kwargs):
        """Crear la List View del modelo"""
        # Class
        mixins = import_all_mixins() + [UpdateMixin]
        if self.site.inlines and isinstance(self.site.inlines, (list, tuple, dict)):
            InlinesMixin = import_mixin("InlinesMixin")

            class Inlines(InlinesMixin):
                inlines = self.site.inlines

            mixins += [Inlines]

        View = get_base_view(BaseUpdateView, mixins, self.get_site())

        # Set attribures
        View.form_class = self.site.form_class
        View.fields = self.site.fields
        if self.site.update_mixins:
            View.__bases__ = (*self.site.update_mixins, *View.__bases__)
        else:
            View.__bases__ = (*self.site.form_mixins, *View.__bases__)
        view = View.as_view()
        return view(request, *args, **kwargs)


class MassUpdateView(View):
    site = None
    http_method_names = [
        "post",
    ]

    def post(self, request, *args, **kwargs):
        model = self.site.model
        ids = request.POST.getlist("ids")
        field = request.POST.get("field")
        value = request.POST.get("value")

        if not self.request.user.has_perm(
            f"{model._meta.app_label}.update_{model._meta.model_name}"
        ):
            return HttpResponseForbidden()

        if not hasattr(model, field):
            return JsonResponse({"error": "El campo indicado no existe."}, status=400)

        objects = model.objects.filter(id__in=ids)
        objects.update(**{field: value})

        return JsonResponse(
            {"success": f"{len(objects)} objectos actualizados."}, status=200
        )



# Models
from .models import Menu
from .utils import get_user_menu


def menu(request):
    return {"menu_tree": build_user_menu(request.user)}


def build_user_menu(user):
    if not user.is_authenticated or not user.is_active:
        return []

    user_perms = set(user.get_all_permissions())
    object_list = (
        Menu.objects.filter(parent__isnull=True, is_active=True)
        .select_related("action", "parent")
        .prefetch_related(
            "submenus",
            "submenus__action",
            "submenus__parent",
            "submenus__submenus",
            "submenus__submenus__action",
            "submenus__submenus__parent",
            "submenus__submenus__submenus",
            "submenus__submenus__submenus__action",
            "submenus__submenus__submenus__parent",
            "submenus__submenus__submenus__submenus",
            "submenus__submenus__submenus__submenus__action",
            "submenus__submenus__submenus__submenus__parent",
            "submenus__submenus__submenus__submenus__submenus",
            "submenus__submenus__submenus__submenus__submenus__action",
            "submenus__submenus__submenus__submenus__submenus__parent",
        )
    )
    return get_user_menu(object_list, user, user_perms)





""" """
# Python
from datetime import datetime

# Django
from django.apps import apps
from django.http import JsonResponse
from django.views.generic import View

# Local
from ..services import FieldService, FilterService


class FilterView(View):
    http_method_names = ["get"]

    def get(self, request, *args, **kwargs):
        model = apps.get_model(self.app_name, self.model_name)
        lookups = self.get_lookups(model, self.field)
        choices = self.get_choices(model, self.field)
        return JsonResponse(
            {
                "lookups": lookups,
                "choices": choices,
                "type": FieldService.get_field_type(model, self.field),
            }
        )

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.app_name = kwargs.get("app")
        self.model_name = kwargs.get("model")
        self.field = kwargs.get("field")

    def get_lookups(self, model, field):
        lookups = [
            {"id": lookup, "text": label}
            for lookup, label in FilterService.get_field_lookups(model, field)
        ]
        return lookups

    def get_choices(self, model, field):
        choices = FilterService.get_choices(model, field)
        if hasattr(choices, "model"):
            choices = [{"id": obj.id, "text": str(obj)} for obj in choices]
        else:
            choices = [{"id": value, "text": label} for value, label in choices]
        return choices


class SessionView(View):
    http_method_names = ["post"]

    def post(self, request, *args, **kwargs):
        params = self.save_params()
        return JsonResponse(params)

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.app_name = kwargs.get("app")
        self.model_name = kwargs.get("model")

    def save_params(self):
        model = apps.get_model(self.app_name, self.model_name)
        params = {}
        for key, value in self.request.POST.items():
            if key != "csrfmiddlewaretoken":
                clean_key = key.split("__")
                if clean_key:
                    clean_key = clean_key[0]
                    field = FieldService.get_field_type(model, clean_key)
                    if field == "DateField":
                        datetime_object = datetime.strptime(value, "%d/%m/%Y").strftime(
                            "%Y-%m-%d"
                        )
                        params.update({key: datetime_object})
                    else:
                        params.update({key: value})
        filters = self.request.session.get("filters", [])
        now = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        for elem in filters:
            if elem["app"] == self.app_name and elem["model"] == self.model_name:
                elem["params"] = params
                elem["last_date"] = now
                break
        else:
            filters.append(
                {
                    "app": self.app_name,
                    "model": self.model_name,
                    "params": params,
                    "last_date": now,
                }
            )
        self.request.session["filters"] = filters
        return params



# Python
from copy import copy
import types
import re

# Django
from django import template
from django.urls import reverse_lazy
from django.forms.models import model_to_dict
from django.contrib.contenttypes.models import ContentType
from django.utils.safestring import mark_safe
from collections.abc import Iterable
from .. import settings


from django.template import Context
from django.template.base import (
    FILTER_SEPARATOR,
    VariableNode,
    TextNode,
    Node,
    NodeList,
    TemplateSyntaxError,
    VariableDoesNotExist,
)

# Exceptions
from django.template.exceptions import TemplateDoesNotExist

from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.forms.utils import pretty_name


register = template.Library()


def silence_without_field(fn):
    """
    Args:
        fn:
    """

    def wrapped(field, attr):
        if not field:
            return ""
        return fn(field, attr)

    return wrapped


def _process_field_attributes(field, attr, process):

    # split attribute name and value from 'attr:value' string
    """
    Args:
        field:
        attr:
        process:
    """
    params = attr.split(":", 1)
    attribute = params[0]
    value = params[1] if len(params) == 2 else ""

    field = copy(field)

    # decorate field.as_widget method with updated attributes
    old_as_widget = field.as_widget

    def as_widget(self, widget=None, attrs=None, only_initial=False):
        attrs = attrs or {}
        process(widget or self.field.widget, attrs, attribute, value)
        html = old_as_widget(widget, attrs, only_initial)
        self.as_widget = old_as_widget
        return html

    field.as_widget = types.MethodType(as_widget, field)
    return field


@register.filter("attr")
@silence_without_field
def set_attr(field, attr):
    """
    Args:
        field:
        attr:
    """

    def process(widget, attrs, attribute, value):
        attrs[attribute] = value

    return _process_field_attributes(field, attr, process)


@register.filter("append_attr")
@silence_without_field
def append_attr(field, attr):
    """
    Args:
        field:
        attr:
    """

    def process_str(field, attr):
        params = attr.split(":", 1)
        attribute = params[0]
        value = params[1] if len(params) == 2 else ""

        content = field.split("name")
        content.insert(1, f'{attribute}="{value}" name')
        field = "".join(content)
        return mark_safe(field)

    if isinstance(field, str):
        return process_str(field, attr)

    def process(widget, attrs, attribute, value):
        if attrs.get(attribute):
            attrs[attribute] += " " + value
        elif widget.attrs.get(attribute):
            attrs[attribute] = widget.attrs[attribute] + " " + value
        else:
            attrs[attribute] = value

    return _process_field_attributes(field, attr, process)


@register.filter("add_class")
@silence_without_field
def add_class(field, css_class):
    """
    Args:
        field:
        css_class:
    """
    return append_attr(field, "class:" + css_class)


@register.filter("data")
@silence_without_field
def set_data(field, data):
    return set_attr(field, "data-" + data)


@register.filter(name="field_type")
def field_type(field):
    """Template filter that returns field class name (in lower case). E.g. if
    field is CharField then {{ field|field_type }} will return 'charfield'.

    Args:
        field:
    """
    if hasattr(field, "field") and field.field:
        return field.field.__class__.__name__.lower()
    return ""


@register.filter(name="widget_name")
def widget_name(field):
    """Template filter that returns field widget class name (in lower case).
    E.g. if field's widget is TextInput then {{ field|widget_type }} will return
    'textinput'.

    Args:
        field:
    """
    if (
        hasattr(field, "field")
        and hasattr(field.field, "widget")
        and field.field.widget
    ):
        return field.field.widget.__class__.__name__.lower()
    return ""


# ======================== render_field tag ==============================

ATTRIBUTE_RE = re.compile(
    r"""
    (?P<attr>
        [\w_-]+
    )
    (?P<sign>
        \+?=
    )
    (?P<value>
    ['"]? # start quote
        [^"']*
    ['"]? # end quote
    )
""",
    re.VERBOSE | re.UNICODE,
)

# ATTRIBUTE_RE = re.compile(r"""(?P<attr>[\w_-]+)(?P<sign>\+?=)(?P<value>['"]?[^"']*['"]?)""", re.VERBOSE | re.UNICODE)


@register.tag
def render_field(parser, token):
    """Render a form field using given attribute-value pairs

    Takes form field as first argument and list of attribute-value pairs for
    all other arguments. Attribute-value pairs should be in the form of
    attribute=value or attribute="a value" for assignment and attribute+=value
    or attribute+="value" for appending.

    Args:
        parser:
        token:
    """
    error_msg = (
        '%r tag requires a form field followed by a list of attributes and values in the form attr="value"'
        % token.split_contents()[0]
    )
    try:
        bits = token.split_contents()
        tag_name = bits[0]
        form_field = bits[1]
        attr_list = bits[2:]
    except ValueError:
        raise TemplateSyntaxError(error_msg)

    form_field = parser.compile_filter(form_field)

    attrs = []
    for pair in attr_list:
        match = ATTRIBUTE_RE.match(pair)
        if not match:
            raise TemplateSyntaxError(error_msg + ": %s" % pair)
        dct = match.groupdict()
        attr, value = dct["attr"], parser.compile_filter(dct["value"])
        attrs.append((attr, value))

    return FieldNode(form_field, attrs)


class FieldNode(Node):
    def __init__(self, field, attrs):
        """
        Args:
            field:
            attrs:
        """
        self.field = field
        self.attrs = attrs

    def render(self, context):
        """
        Args:
            context:
        """

        bounded_field = self.field.resolve(context)
        # field = getattr(bounded_field, "field", None)
        with context.push():
            for key, value in self.attrs:
                if key == "class":
                    bounded_field = add_class(bounded_field, value.resolve(context))
                elif key == "type":
                    bounded_field.field.widget.input_type = value.resolve(context)
                else:
                    bounded_field = set_attr(
                        bounded_field, "%s:%s" % (key, value.resolve(context))
                    )
                    context.update({key: value.resolve(context)})

            # Get template name
            if not bounded_field:
                raise ImproperlyConfigured("The field passed do not exist.")

            widget_name = bounded_field.widget_type
            if not widget_name in settings.TEMPLATE_WIDGETS:
                if not "default" in settings.TEMPLATE_WIDGETS:
                    raise ImproperlyConfigured(
                        f"Does not exist template name for '{widget_name}' or default widget template."
                    )
                template_name = settings.TEMPLATE_WIDGETS.get("default")
            else:
                template_name = settings.TEMPLATE_WIDGETS.get(widget_name)

            context.update(
                {
                    "field": bounded_field,
                }
            )

            if (
                widget_name in ("clearablefile",)
                and hasattr(bounded_field.form, "instance")
                and hasattr(bounded_field.form.instance, bounded_field.name)
            ):
                file = getattr(bounded_field.form.instance, bounded_field.name)
                if file:
                    context.update({"file_name": file.name, "file_url": file.url})

            t = context.template.engine.get_template(template_name)
            component = t.render(context)
            context.pop()
            return mark_safe(component)


"""
@register.filter
def verbose_name(obj):
    
    return obj._meta.verbose_name


@register.filter
def verbose_name_plural(obj):
   
    try:
        return obj._meta.verbose_name_plural
    except:
        pass


@register.simple_tag
def get_verbose_field_name(instance, field_name):
   
    return instance._meta.get_field(field_name).verbose_name.title()


@register.filter
def get_class(obj):
    
    return obj.model.__name__


# OBTENER LOS NOMBRES DE UN OBJETO
# GET MODEL INSTANCE FIELDS
# EJ. OBJECT.OBJECTS.GET(PK=1)
@register.filter
def get_field_names(obj):
    
    return obj._meta.fields
    # El siguiente metodo recupera incluso las tablas subrelacionadas
    # return obj._meta.get_fields()


@register.filter
def get_field_verbose_names(obj):
   
    return obj._meta.verbose_name.title()


@register.simple_tag
def get_field_values(obj):
    
    return model_to_dict(obj)
"""

# LIST VIEW CON VALUES
"""
@register.simple_tag
def obtener_cabeceras(obj, numero):
    
    objeto = model_to_dict(obj)
    cabeceras = []
    for c in objeto.keys():
        cabeceras.append(obj._meta.get_field(c).verbose_name.title())
    return cabeceras


@register.simple_tag
def obtener_valores(obj, numero):
   
    objeto = model_to_dict(obj)
    objeto = model_to_dict(obj)
    valores = []
    for c in objeto.values():
        valores.append(c)
    return valores


@register.filter
def verbose_name_and_name(obj):
   
    objeto = model_to_dict(obj)
    lista = []
    for c in objeto.keys():
        lista.append(
            {"verbose_name": obj._meta.get_field(c).verbose_name.title(), "name": c}
        )
    return lista








@register.filter
def is_date(val):
  
    return isinstance(val, date)


@register.simple_tag
def get_headers(model_site):
 
    headers = []
    for name in model_site.list_display:
        try:
            field = model_site.model._meta.get_field(name)
            label = field.verbose_name
        except FieldDoesNotExist:
            if name == "__str__":
                label = str(model_site.model._meta.verbose_name)
            else:
                if hasattr(model_site, name):
                    attr = getattr(model_site, name)
                elif hasattr(model_site.model, name):
                    attr = getattr(model_site.model, name)
                else:
                    raise AttributeError

                label = attr.__name__ if callable(attr) else attr
        finally:
            headers.append(pretty_name(label))

    return headers


@register.simple_tag
def get_results(model_site, queryset):
    results = []
    for instance in queryset:
        try:
            slug_or_pk = instance.slug
        except:
            slug_or_pk = instance.id
        line = {
            "update_url": reverse_lazy(
                "site:%s_%s_editar" % model_site.get_info(), args=[instance.id]
            ),
            "detail_url": reverse_lazy(
                "site:%s_%s_detalle" % model_site.get_info(), args=[instance.id]
            ),
        }
        object_list = []
        for name in model_site.list_display:
            if name == "__str__":
                value = str(instance)
            else:
                if hasattr(model_site, name):
                    attr = getattr(model_site, name)
                elif hasattr(instance, name):
                    attr = getattr(instance, name)
                else:
                    raise AttributeError
                value = attr() if callable(attr) else attr
            object_list.append(value)
        line.update({"object_list": object_list})
        results.append(line)
    return results



@register.simple_tag
def model_name(value):
    if hasattr(value, "model"):
        value = value.model

    return value._meta.label.split(".")[-1]


@register.filter
def content_type(obj):
    if not obj:
        return False
    return ContentType.objects.get_for_model(obj).id


@register.filter()
def get_app_label_content_type(value):
    return ContentType.objects.get_for_id(value)


@register.filter
def has_permission(request, view_permissions):
    user = request.user
    if request.user.is_authenticated and request.user.is_superuser:
        return True
    permissions = list()
    if isinstance(view_permissions, Iterable):
        for model in view_permissions:
            permissions.append(f"{model._meta.app_label}.view_{model._meta.model_name}")
            permissions.append(f"{model._meta.app_label}.add_{model._meta.model_name}")
            permissions.append(
                f"{model._meta.app_label}.change_{model._meta.model_name}"
            )
    else:
        permissions.append(
            f"{view_permissions._meta.app_label}.view_{view_permissions._meta.model_name}"
        )
        permissions.append(
            f"{view_permissions._meta.app_label}.add_{view_permissions._meta.model_name}"
        )
        permissions.append(
            f"{view_permissions._meta.app_label}.change_{view_permissions._meta.model_name}"
        )
    return any(user.has_perm(permission) for permission in permissions)


@register.simple_tag
def permission(request, object, action):
    try:
        model = object.__class__
    except:
        model = object
    return request.user.has_perm(
        f"{model._meta.app_label}.{action}_{model._meta.model_name}"
    )


@register.filter
def get_type(value):
    return type(value)


# RENDERIZAR PDF
@register.simple_tag(takes_context=True)
def render_pdf(context, tpl_string):
    t = template.Template(tpl_string)
    return t.render(context)


@register.simple_tag(takes_context=True)
def params(context, **kwargs):
    
    Return encoded URL parameters that are the same as the current
    request's parameters, only with the specified GET parameters added or changed.

    It also removes any empty parameters to keep things neat,
    so you can remove a parm by setting it to ``""``.

    For example, if you're on the page ``/things/?with_frosting=true&page=5``,
    then

    <a href="/things/?{% params page=3 %}">Page 3</a>

    would expand to

    <a href="/things/?with_frosting=true&page=3">Page 3</a>
    d = context["request"].GET.copy()
    for k, v in kwargs.items():
        d[k] = v
    for k in [k for k, v in d.items() if not v]:
        del d[k]
    return d.urlencode()
    
"""



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gmcm-django-superadmin",
    version="2.1.0",
    author="Denis Siavichay",
    author_email="dbsiavichay@gmail.com",
    description="A Django app for build admin sites like django admin.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/mmorona/gmcm-django-superadmin",
    packages=setuptools.find_packages(),
    install_requires=[
        "pyyaml",
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)
